@@ -24,12 +24,11 @@
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
-
-
 import logging
-from typing import Dict, List, Optional, Tuple, Union
+from typing import Dict, Iterable, List, Optional, Tuple, Union
 
 import numpy as np
+import pandas as pd
 from component_contribution import DEFAULT_QUILT_PKG as DEFAULT_QUILT_PKG_CC
 from component_contribution import (
     DEFAULT_QUILT_VERSION as DEFAULT_QUILT_VERSION_CC,
@@ -52,16 +51,19 @@ from equilibrator_cache.api import DEFAULT_QUILT_PKG as DEFAULT_QUILT_PKG_CACHE
 from equilibrator_cache.api import (
     DEFAULT_QUILT_VERSION as DEFAULT_QUILT_VERSION_CACHE,
 )
+from equilibrator_cache.reaction import (
+    create_stoichiometric_matrix_from_reactions,
+)
 
 from . import (
     FARADAY,
     Q_,
     R,
-    default_I,
-    default_pH,
-    default_pMg,
+    default_physiological_ionic_strength,
+    default_physiological_p_h,
+    default_physiological_p_mg,
+    default_physiological_temperature,
     default_rmse_inf,
-    default_T,
     ureg,
 )
 from .phased_reaction import PhasedReaction
@@ -72,38 +74,44 @@ logger = logging.getLogger(__name__)
 
 
 def find_most_abundance_ms(
-    cpd: Compound, p_h: Q_, ionic_strength: Q_, temperature: Q_
+    cpd: Compound, p_h: Q_, p_mg: Q_, ionic_strength: Q_, temperature: Q_
 ) -> CompoundMicrospecies:
-    """Find the most abundant microspecies based on transformed energies.
-
-    Ignore microspecies with Magnesium, as we assume they never traverse
-    the membrane. Therefore, the value of p_mg is not necessary.
-    """
-    p_mg = default_pMg
+    """Find the most abundant microspecies based on transformed energies."""
     ddg_over_rts = [
-        (ms.transform(p_h, ionic_strength, temperature, p_mg).magnitude, ms)
+        (
+            ms.transform(
+                pH=p_h.m_as(""),
+                pMg=p_mg.m_as(""),
+                ionic_strength_M=ionic_strength.m_as("M"),
+                T_in_K=temperature.m_as("K"),
+            ),
+            ms,
+        )
         for ms in cpd.microspecies
-        if ms.number_magnesiums == 0
     ]
     min_ddg, min_ms = min(ddg_over_rts, key=lambda x: x[0])
     return min_ms
 
 
 def predict_protons_and_charge(
-    rxn: PhasedReaction, p_h: Q_, ionic_strength: Q_, temperature: Q_
-) -> Tuple[float, float]:
+    rxn: PhasedReaction, p_h: Q_, p_mg: Q_, ionic_strength: Q_, temperature: Q_
+) -> Tuple[float, float, float]:
     """Find the #protons and charge of a transport half-reaction."""
     n_h = 0
+    n_mg = 0
     z = 0
     for cpd, coeff in rxn.items():
         if cpd.inchi_key == PROTON_INCHI_KEY:
             n_h += coeff
             z += coeff
         else:
-            ms = find_most_abundance_ms(cpd, p_h, ionic_strength, temperature)
+            ms = find_most_abundance_ms(
+                cpd, p_h, p_mg, ionic_strength, temperature
+            )
             n_h += coeff * ms.number_protons
+            n_mg += coeff * ms.number_magnesiums
             z += coeff * ms.charge
-    return n_h, z
+    return n_h, n_mg, z
 
 
 class ComponentContribution(object):
@@ -129,10 +137,10 @@ class ComponentContribution(object):
         ccache : CompoundCache, optional
         predictor : GibbsEnergyPredictor, optional
         """
-        self._p_h = default_pH
-        self._ionic_strength = default_I
-        self._p_mg = default_pMg
-        self._temperature = default_T
+        self._p_h = default_physiological_p_h
+        self._ionic_strength = default_physiological_ionic_strength
+        self._p_mg = default_physiological_p_mg
+        self._temperature = default_physiological_temperature
 
         if rmse_inf is not None:
             assert rmse_inf.check(
@@ -143,7 +151,7 @@ class ComponentContribution(object):
 
         if predictor is None:
             parameters = CCModelParameters.from_quilt(
-                version=LEGACY_QUILT_VERSION_CC, force=True
+                version=DEFAULT_QUILT_VERSION_CC, force=True
             )
             self.predictor = GibbsEnergyPredictor(
                 parameters=parameters, rmse_inf=rmse_inf
@@ -200,6 +208,25 @@ class ComponentContribution(object):
         ), "temperature must be in appropriate units (e.g. Kalvin)"
         self._temperature = value
 
+    @staticmethod
+    def legacy() -> "ComponentContribution":
+        """Initialize a ComponentContribution object with the legacy version.
+
+        The legacy version is intended for compatibility with older versions
+        of equilibrator api (0.2.x - 0.3.1).
+        Starting from 0.3.2, there is a significant change in the predictions
+        caused by an improved Mg2+ concentration model.
+
+        Returns
+        -------
+        A ComponentContribution object
+        """
+        cc = ComponentContribution.initialize_custom_version(
+            quilt_version_cc=LEGACY_QUILT_VERSION_CC
+        )
+        cc.p_mg = Q_(14)  # i.e. set [Mg2+] to (essentially) zero
+        return cc
+
     @staticmethod
     def initialize_custom_version(
         rmse_inf: Q_ = default_rmse_inf,
@@ -224,13 +251,13 @@ class ComponentContribution(object):
         quilt_package_cache : str, optional
             (Default value: "equilibrator/cache")
         quilt_version_cache : str, optional
-            (Default value: "0.2.7")
+            (Default value: "0.2.10")
         quilt_tag_cache : str, optional
         quilt_hash_cache: str, optional
         quilt_package_cc : str, optional
             (Default value: "equilibrator/component_contribution")
         quilt_version_cc : str, optional
-            (Default value: "0.2.18")
+            (Default value: "0.3.0")
         quilt_tag_cc : str, optional
         quilt_hash_cc: str, optional
 
@@ -261,29 +288,42 @@ class ComponentContribution(object):
     def get_compound(self, compound_id: str) -> Union[Compound, None]:
         """Get a Compound using the DB namespace and its accession.
 
-        :return: a Compound object or None
+        Returns
+        -------
+        cpd : Compound
         """
         return self.ccache.get_compound(compound_id)
 
     def get_compound_by_inchi(self, inchi: str) -> Union[Compound, None]:
         """Get a Compound using InChI.
 
-        :return: a Compound object or None
+        Returns
+        -------
+        cpd : Compound
         """
         return self.ccache.get_compound_by_inchi(inchi)
 
     def search_compound_by_inchi_key(self, inchi_key: str) -> List[Compound]:
         """Get a Compound using InChI.
 
-        :return: a Compound object or None
+        Returns
+        -------
+        cpd : Compound
         """
         return self.ccache.search_compound_by_inchi_key(inchi_key)
 
     def search_compound(self, query: str) -> Union[None, Compound]:
         """Try to find the compound that matches the name best.
 
-        :param query: a compound name
-        :return: a Compound object of the best match
+        Parameters
+        ----------
+        query : str
+            an (approximate) compound name
+
+        Returns
+        -------
+        cpd : Compound
+            the best match
         """
         hits = self.ccache.search(query)
         if not hits:
@@ -303,16 +343,28 @@ class ComponentContribution(object):
     def parse_reaction_formula(self, formula: str) -> PhasedReaction:
         """Parse reaction text using exact match.
 
-        :param formula: a string containing the reaction formula
-        :return: a Reaction object
+        Parameters
+        ----------
+        formula : str
+            a string containing the reaction formula
+
+        Returns
+        -------
+        rxn : Reaction
         """
         return PhasedReaction.parse_formula(self.ccache.get_compound, formula)
 
     def search_reaction(self, formula: str) -> PhasedReaction:
-        """Parse a reaction written using compound names and approximate matching.
+        """Search a reaction written using compound names (approximately).
 
-        :param formula: a string containing the reaction formula
-        :return: a Reaction object
+        Parameters
+        ----------
+        formula : str
+            a string containing the reaction formula
+
+        Returns
+        -------
+        rxn : Reaction
         """
         results = self.reaction_parser.parseString(formula)
         substrates, arrow, products = results
@@ -391,13 +443,34 @@ class ComponentContribution(object):
         """Get the value of RT."""
         return R * self.temperature
 
+    def standard_dg_formation(
+        self, compound: Compound
+    ) -> Tuple[Optional[float], Optional[np.ndarray]]:
+        """Get the (mu, sigma) predictions of a compound's formation energy.
+
+        Parameters
+        ----------
+        compound: Compound
+            a Compound object
+
+        Returns
+        -------
+        mu : float
+            the mean of the standard Gibbs energy of formation
+        sigma : array
+            the uncertainty vector
+
+        """
+        return self.predictor.get_compound_prediction(compound)
+
     def standard_dg(self, reaction: PhasedReaction) -> ureg.Measurement:
         """Calculate the chemical reaction energies of a reaction.
 
-        :param reaction: the input Reaction object
-        :return: a tuple of the dG0 in kJ/mol and standard error. to
-        calculate the confidence interval, use the range -1.96 to 1.96 times
-        this value
+        Returns
+        -------
+        standard_dg : ureg.Measurement
+            the dG0 in kJ/mol and standard error. To calculate the 95%
+            confidence interval, use the range -1.96 to 1.96 times this value
         """
         residual_reaction, stored_dg = reaction.separate_stored_dg()
 
@@ -407,13 +480,18 @@ class ComponentContribution(object):
 
     def standard_dg_multi(
         self, reactions: List[PhasedReaction]
-    ) -> Tuple[np.array, np.array]:
+    ) -> Tuple[np.ndarray, np.ndarray]:
         """Calculate the chemical reaction energies of a list of reactions.
 
         Using the major microspecies of each of the reactants.
 
-        :return: a tuple with the CC estimation of the major microspecies'
-        standard formation energy, and the uncertainty matrix.
+        Returns
+        -------
+        standard_dg : ndarray
+            the CC estimation of the major microspecies' standard formation
+            energy
+        dg_sigma : ndarray
+            the uncertainty matrix.
         """
         rxn_dg_pairs = map(lambda r: r.separate_stored_dg(), reactions)
         residual_reactions, stored_dg = zip(*rxn_dg_pairs)
@@ -427,10 +505,11 @@ class ComponentContribution(object):
     def standard_dg_prime(self, reaction: PhasedReaction) -> ureg.Measurement:
         """Calculate the transformed reaction energies of a reaction.
 
-        :param reaction: the input Reaction object
-        :return: a tuple of the dG0_prime in kJ/mol and standard error. to
-        calculate the confidence interval, use the range -1.96 to 1.96 times
-        this value
+        Returns
+        -------
+        standard_dg : ureg.Measurement
+            the dG0_prime in kJ/mol and standard error. To calculate the 95%
+            confidence interval, use the range -1.96 to 1.96 times this value
         """
         residual_reaction, stored_dg_prime = reaction.separate_stored_dg_prime(
             p_h=self.p_h,
@@ -452,11 +531,11 @@ class ComponentContribution(object):
     def dg_prime(self, reaction: PhasedReaction) -> ureg.Measurement:
         """Calculate the dG'0 of a single reaction.
 
-        :param reaction: an object of type Reaction
-        :return: a tuple of (dG_r_prime, dG_uncertainty) where dG_r_prime is
-        the estimated Gibbs free energy of reaction and dG_uncertainty is the
-        standard deviation of estimation. Multiply it by 1.96 to get a 95%
-        confidence interval (which is the value shown on eQuilibrator).
+        Returns
+        -------
+        dg : ureg.Measurement
+            the dG_prime in kJ/mol and standard error. To calculate the 95%
+            confidence interval, use the range -1.96 to 1.96 times this value
         """
         return (
             self.standard_dg_prime(reaction)
@@ -468,13 +547,17 @@ class ComponentContribution(object):
     ) -> Tuple[np.array, np.array]:
         """Calculate the transformed reaction energies of a list of reactions.
 
-        :return: a tuple of two arrays. the first is a 1D NumPy array
-        containing the CC estimates for the reactions' transformed dG0
-        the second is a 2D numpy array containing the covariance matrix
-        of the standard errors of the estimates. one can use the eigenvectors
-        of the matrix to define a confidence high-dimensional space, or use
-        U as the covariance of a Gaussian used for sampling
-        (where dG0_cc is the mean of that Gaussian).
+        Returns
+        -------
+        standard_dg_primes : ndarray
+            a 1D NumPy array containing the CC estimates for the reactions'
+            transformed dG0
+        dg_sigma : ndarray
+            the second is a 2D numpy array containing the covariance matrix
+            of the standard errors of the estimates. one can use the
+            eigenvectors of the matrix to define a confidence high-dimensional
+            space, or use dg_sigma as the covariance of a Gaussian used for
+            sampling (where 'standard_dg_primes' is the mean of that Gaussian).
         """
         rxn_dg_pairs = map(
             lambda r: r.separate_stored_dg_prime(
@@ -505,11 +588,17 @@ class ComponentContribution(object):
         Assume all aqueous reactants are at 1 mM, gas reactants at 1 mbar and
         the rest at their standard concentration.
 
-        :param reaction: an object of type PhasedReaction
-        :return: a tuple (dG_r_prime, dG_uncertainty) where dG_r_prime is
-        the estimated Gibbs free energy of reaction and dG_uncertainty is the
-        standard deviation of estimation. Multiply it by 1.96 to get a 95%
-        confidence interval (which is the value shown on eQuilibrator).
+        Returns
+        -------
+        standard_dg_primes : ndarray
+            a 1D NumPy array containing the CC estimates for the reactions'
+            physiological dG'
+        dg_sigma : ndarray
+            the second is a 2D numpy array containing the covariance matrix
+            of the standard errors of the estimates. one can use the
+            eigenvectors of the matrix to define a confidence high-dimensional
+            space, or use dg_sigma as the covariance of a Gaussian used for
+            sampling (where 'standard_dg_primes' is the mean of that Gaussian).
         """
         return (
             self.standard_dg_prime(reaction)
@@ -521,7 +610,10 @@ class ComponentContribution(object):
     ) -> ureg.Measurement:
         """Calculate the reversibility index (ln Gamma) of a single reaction.
 
-        :return: ln_RI - The reversibility index (in natural log scale).
+        Returns
+        -------
+        ln_RI : ureg.Measurement
+            the reversibility index (in natural log scale).
         """
         physiological_dg_prime = self.physiological_dg_prime(reaction)
 
@@ -534,8 +626,9 @@ class ComponentContribution(object):
     def standard_e_prime(self, reaction: PhasedReaction) -> ureg.Measurement:
         """Calculate the E'0 of a single half-reaction.
 
-        :param reaction: a PhasedReaction object
-        :return: a tuple (E0_prime, E0_uncertainty) where E0_prime is
+        Returns
+        -------
+        standard_e_prime : ureg.Measurement
         the estimated standard electrostatic potential of reaction and
         E0_uncertainty is the standard deviation of estimation. Multiply it
         by 1.96 to get a 95% confidence interval (which is the value shown on
@@ -548,16 +641,18 @@ class ComponentContribution(object):
             raise ValueError(
                 "this is not a half-reaction, " "electrons are balanced"
             )
-        return -self.standard_dg_prime(reaction) / (n_e * FARADAY)
+        standard_e_prime = -self.standard_dg_prime(reaction) / (n_e * FARADAY)
+        return standard_e_prime.to("mV")
 
     def physiological_e_prime(
         self, reaction: PhasedReaction
     ) -> ureg.Measurement:
         """Calculate the E'0 of a single half-reaction.
 
-        :param reaction: a PhasedReaction object
-        :return: a tuple (E0_prime, E0_uncertainty) where E0_prime is
-        the estimated standard electrostatic potential of reaction and
+        Returns
+        -------
+        physiological_e_prime : ureg.Measurement
+        the estimated physiological electrostatic potential of reaction and
         E0_uncertainty is the standard deviation of estimation. Multiply it
         by 1.96 to get a 95% confidence interval (which is the value shown on
         eQuilibrator).
@@ -574,9 +669,10 @@ class ComponentContribution(object):
     def e_prime(self, reaction: PhasedReaction) -> ureg.Measurement:
         """Calculate the E'0 of a single half-reaction.
 
-        :param reaction: a PhasedReaction object
-        :return: a tuple (E0_prime, E0_uncertainty) where E0_prime is
-        the estimated standard electrostatic potential of reaction and
+        Returns
+        -------
+        e_prime : ureg.Measurement
+        the estimated electrostatic potential of reaction and
         E0_uncertainty is the standard deviation of estimation. Multiply it
         by 1.96 to get a 95% confidence interval (which is the value shown on
         eQuilibrator).
@@ -593,16 +689,18 @@ class ComponentContribution(object):
     def dg_analysis(self, reaction: PhasedReaction) -> List[Dict[str, object]]:
         """Get the analysis of the component contribution estimation process.
 
-        :param reaction: a PhasedReaction object.
-        :return: the analysis results as a list of dictionaries
+        Returns
+        ------
+        the analysis results as a list of dictionaries
         """
         return self.predictor.get_dg_analysis(reaction)
 
     def is_using_group_contribution(self, reaction: PhasedReaction) -> bool:
         """Check whether group contribution is needed to get this reactions' dG.
 
-        :param reaction: a PhasedReaction object.
-        :return: true iff group contribution is needed
+        Returns
+        -------
+        true iff group contribution is needed
         """
         return self.predictor.is_using_group_contribution(reaction)
 
@@ -613,28 +711,58 @@ class ComponentContribution(object):
         delta_chi: Q_,
         p_h_outer: Q_,
         ionic_strength_outer: Q_,
-        p_mg_outer: Q_ = default_pMg,
+        p_mg_outer: Q_ = default_physiological_p_mg,
     ) -> ureg.Measurement:
         """Calculate the transformed energies of a multi-compartmental reaction.
 
         Based on the equations from
         Harandsdottir et al. 2012 (https://doi.org/10.1016/j.bpj.2012.02.032)
 
-        :param reaction_inner: the inner compartment half-reaction
-        :param reaction_outer: the outer compartment half-reaction
-        :param delta_phi: the difference in electro-static potential between
-        the outer and inner compartments
-        :param p_h_outer: the pH in the outside compartment
-        :param ionic_strength_outer: the ionic strength outside
-        :param p_mg_outer: the pMg in the outside compartment
-        :return: the transport reaction Gibbs free energy change
+        Parameters
+        ----------
+        reaction_inner : PhasedReaction
+            the inner compartment half-reaction
+        reaction_outer : PhasedReaction
+            the outer compartment half-reaction
+        delta_phi : Q_
+            the difference in electro-static potential between
+            the outer and inner compartments
+        p_h_outer : Q_
+            the pH in the outside compartment
+        ionic_strength_outer : Q_
+            the ionic strength outside
+        p_mg_outer : Q_
+            the pMg in the outside compartment
+
+        Returns
+        -------
+        standard_dg_prime : ureg.Measurement
+            the transport reaction Gibbs free energy change
         """
-        n_h_inner, z_inner = predict_protons_and_charge(
-            reaction_inner, self.p_h, self.ionic_strength, self.temperature
+
+        # We want to ignore microspecies with Magnesium, as we assume they
+        # never traverse the membrane.
+        # Therefore, we set the value of p_mg is set to 14 (very high), so that
+        # species with Mg2+ would never be the most abundant.
+        p_mg = Q_(14)
+
+        n_h_inner, n_mg_inner, z_inner = predict_protons_and_charge(
+            reaction_inner,
+            self.p_h,
+            p_mg,
+            self.ionic_strength,
+            self.temperature,
         )
-        n_h_outer, z_outer = predict_protons_and_charge(
-            reaction_outer, self.p_h, self.ionic_strength, self.temperature
+        n_h_outer, n_mg_outer, z_outer = predict_protons_and_charge(
+            reaction_outer,
+            self.p_h,
+            p_mg,
+            self.ionic_strength,
+            self.temperature,
         )
+        assert n_mg_inner == 0, "most abundant inner species have Mg2+"
+        assert n_mg_outer == 0, "most abundant outer species have Mg2+"
+
         if (n_h_inner != -n_h_outer) or (z_inner != -z_outer):
             raise ValueError(
                 "inner and outer half-reactions don't balance each other: "
@@ -679,3 +807,29 @@ class ComponentContribution(object):
         )
 
         return standard_dg + stored_dg_prime_inner + stored_dg_prime_outer
+
+    def create_stoichiometric_matrix(
+        self, reactions: Iterable[PhasedReaction]
+    ) -> pd.DataFrame:
+        """Build a stoichiometric matrix.
+
+        Parameters
+        ----------
+        reactions : Iterable[Reaction]
+            The collection of reactions to build a stoichiometric
+            matrix from.
+
+        Returns
+        -------
+        DataFrame
+            The stoichiometric matrix as a DataFrame whose indexes are the
+            compounds and its columns are the reactions (in the same order as
+            the input).
+
+        """
+        return create_stoichiometric_matrix_from_reactions(
+            reactions,
+            self.ccache.is_proton,
+            self.ccache.is_water,
+            self.ccache.water,
+        )
